//////////////////////////////////////////////////////////////////////////
/// @file   app.h
//////////////////////////////////////////////////////////////////////////

#ifndef _CX_APP_H_
#define _CX_APP_H_

#include "cx/build.h"

// ----------------------------------------------------------------------------
// typedefs
// ----------------------------------------------------------------------------
// Do includes first to insure that even wxBase consistently uses
// the wxMSW case for defining wxAppInitializerFunction.
// (otherwise, __WXMSW__ will not yet be set, if app.h is the
//  first file to be included).

class cxApp;
typedef cxApp* (*cxAppInitializerFunction)();

class cxAppBase
{
public:
    cxAppBase();
    ~cxAppBase();

    // called during the program initialization, returning FALSE from here
    // prevents the program from continuing - it's a good place to create
    // the top level program window and return TRUE.
    //
    // Override: always in GUI application, rarely in console ones.
    virtual bool OnInit();

    // a platform-dependent version of OnInit(): the code here is likely to
    // depend on the toolkit. default version does nothing.
    //
    // Override: rarely.
    bool OnInitGui();

    // called to start program execution - the default version just enters
    // the main GUI loop in which events are received and processed until
    // the last window is not deleted (if GetExitOnFrameDelete) or
    // ExitMainLoop() is called. In console mode programs, the execution
    // of the program really starts here
    //
    // Override: rarely in GUI applications, always in console ones.
    int OnRun();

    // execute the main GUI loop, the function returns when the loop ends
    virtual int MainLoop() = 0;


    // check that the wxBuildOptions object (constructed in the application
    // itself, usually the one from IMPLEMENT_APP() macro) matches the build
    // options of the library and abort if it doesn't
    static bool CheckBuildOptions(const cxBuildOptions& buildOptions);

    // implementation only from now on
    // -------------------------------

    // helpers for dynamic cxApp construction
    static void SetInitializerFunction(cxAppInitializerFunction fn)
        { m_appInitFn = fn; }
    static cxAppInitializerFunction GetInitializerFunction()
        { return m_appInitFn; }
    
protected:
    // function used for dynamic cxApp creation
    static cxAppInitializerFunction m_appInitFn;
    
};

// ----------------------------------------------------------------------------
// now include the declaration of the real class
// ----------------------------------------------------------------------------

#include "cx/msw/app.h"

// ----------------------------------------------------------------------------
// the global data
// ----------------------------------------------------------------------------
// the one and only application object - use of wxTheApp in application code
// is discouraged, consider using DECLARE_APP() after which you may call
// wxGetApp() which will return the object of the correct type (i.e. MyApp and
// not wxApp)
CXDLLEXPORT_DATA(extern cxApp*) cxTheApp;

// ----------------------------------------------------------------------------
// macros for dynamic creation of the application object
// ----------------------------------------------------------------------------

// Having a global instance of this class allows cxApp to be aware of the app
// creator function. cxApp can then call this function to create a new app
// object. Convoluted, but necessary.

class CXDLLEXPORT cxAppInitializer
{
public:
    cxAppInitializer(cxAppInitializerFunction fn)
        { cxApp::SetInitializerFunction(fn); }
};

#define IMPLEMENT_CXWIN_MAIN

// Use this macro if you want to define your own main() or WinMain() function
// and call wxEntry() from there.
#define IMPLEMENT_APP_NO_MAIN(appname)                   \
    cxApp *cxCreateApp()                                 \
    {                                                    \
        cxApp::CheckBuildOptions(cxBuildOptions());      \
        return new appname;                              \
    }                                                    \
    cxAppInitializer cxTheAppInitializer((cxAppInitializerFunction) cxCreateApp); \
    appname& cxGetApp() { return *(appname *)cxTheApp; }



// Use this macro exactly once, the argument is the name of the wxApp-derived
// class which is the class of your application.
#define IMPLEMENT_APP(appname)            \
    IMPLEMENT_APP_NO_MAIN(appname)        \
    IMPLEMENT_CXWIN_MAIN

#endif
    // _CX_APP_H_
