/* inftrees.h -- header to use inftrees.c
 * Copyright (C) 1995-1998 Mark Adler
 * For conditions of distribution and use, see copyright notice in zlib.h 
 */

/* WARNING: this file should *not* be used by applications. It is
   part of the implementation of the compression library and is
   subject to change. Applications should only use zlib.h.
 */

/* Huffman code lookup table entry--this entry is four bytes for machines
   that have 16-bit pointers (e.g. PC's in the small or medium model). */

typedef struct inflate_huft_s inflate_huft;

struct inflate_huft_s {
  union {
    struct {
      z_byte Exop;        /* number of extra bits or operation */
      z_byte Bits;        /* number of bits in this code or subcode */
    } what;
    z_ushort pad;         /* pad structure to a power of 2 (4 bytes for */
  } word;                 /*  16-bit, 8 bytes for 32-bit int's) */
  z_ushort base;          /* literal, length base, distance base,
                           or table offset */
};

/* Maximum size of dynamic tree.  The maximum found in a long but non-
   exhaustive search was 1004 huft structures (850 for length/literals
   and 154 for distances, the latter actually the result of an
   exhaustive search).  The actual maximum is not known, but the
   value below is more than safe. */
#define MANY 1440

extern int inflate_trees_bits (
    z_ushort *,                /* 19 code lengths */
    z_ushort *,                /* bits tree desired/actual depth */
    inflate_huft **,           /* bits tree result */
    inflate_huft *,            /* space for trees */
    z_streamp);                /* for messages */

extern int inflate_trees_dynamic (
    z_ushort,                  /* number of literal/length codes */
    z_ushort,                  /* number of distance codes */
    z_ushort *,                /* that many (total) code lengths */
    z_ushort *,                /* literal desired/actual bit depth */
    z_ushort *,                /* distance desired/actual bit depth */
    inflate_huft **,           /* literal/length tree result */
    inflate_huft **,           /* distance tree result */
    inflate_huft *,            /* space for trees */
    z_streamp);                /* for messages */

extern int inflate_trees_fixed (
    z_ushort *,                /* literal desired/actual bit depth */
    z_ushort *,                /* distance desired/actual bit depth */
    inflate_huft **,           /* literal/length tree result */
    inflate_huft **,           /* distance tree result */
    z_streamp);                /* for memory allocation */
