#include <windows.h>

#define APP_PRIVATE_DECLARATIONS
#include "mg.h"
#include "appcommon.h"
#include "private.h"

char * app_to_native_path(const char *filepath);
void app_set_app_instance(App *app, HINSTANCE instance);

App *app_new_app(int argc, char *argv[])
{
	App *app;
	char *path;

	/* Create the App structure */
	app = app_zero_alloc(sizeof(struct _mgApp));
	if (app == NULL)
		return NULL;
	app->extra = app_zero_alloc(sizeof(struct AppExtra));
	if (app->extra == NULL) {
		app_free(app);
		return NULL;
	}

	/* Copy the program name from argv[0] */
	if (argv && argv[0]) {
		path = app_to_portable_path(argv[0]);
		if ((path[0] != '/') && (path[1] != ':')) {
			app->program_name = app_form_file_path(
					app_current_folder(), path);
		}
		else {
			app->program_name = app_form_file_path("", path);
		}
		app_free(path);
		app->has_resources = app_file_has_resources(app->program_name);
	}
	else {
		app->program_name = app_copy_string("");
		app->has_resources = 0; /* no resources, since no name */
	}

	/* Deactivate X-Windows style mouse-based copy/paste */
	app->use_X_copy_paste = 0;

	app_set_app_instance(app, mghInstance);

	return app;
	return app;
}

void app_del_app(App *app)
{
	app_app_deinitialise(app);
	app_free(app_extra(app));
	app_free(app);
}

void app_set_app_instance(App *app, HINSTANCE instance)
{
	HDC screen;

	/* Initialise string table and similar portable things. */
	app_app_initialise(app);

	/* Is the GUI available? */
	if (instance == 0)
	{
		app->gui_available = 0;
		return;
	}

	app->gui_available = 1;
	app_extra(app)->this_instance = instance;

	/* Initialise the screen dimensions */
	app->screen_area.x = 0;
	app->screen_area.y = 0;
	app->screen_area.width  = GetSystemMetrics(SM_CXSCREEN);
	app->screen_area.height = GetSystemMetrics(SM_CYSCREEN);

	screen = CreateIC("DISPLAY", NULL, NULL, NULL);
	app->screen_mm.x = 0;
	app->screen_mm.y = 0;
	app->screen_mm.width  = GetDeviceCaps(screen, HORZSIZE);
	app->screen_mm.height = GetDeviceCaps(screen, VERTSIZE);
}

int app_exec(App *app, const char *cmd)
{
	char *command;
	int result;

	command = app_to_native_path(cmd);
	if (WinExec(command, SW_SHOW) < 32)
		result = 0;
	else
		result = 1;
	app_free(command);
	return result;
}

void app_beep(App *app)
{
	MessageBeep(-1);
}
