/** 
 *****************************************************************************
 * @file    mg_msw_main.c
 *
 * @brief   The WinMain function -- calls your program's main() function
 *
 *****************************************************************************/

#include <malloc.h>			/* For _alloca() */

#include <windows.h>

#include "mg_main.h"

/* 
 * ----------------------------------------------------------------------------
 * globals
 * ----------------------------------------------------------------------------
 */

/*
 *  Global write-once variable:
 */
HINSTANCE mghInstance = 0;


/*
 *  Routines:
 */


/* Parse a command line buffer into arguments */
static int ParseCommandLine__(char *cmdline, char **argv)
{
	char *bufp;
	int argc;

	argc = 0;
	for ( bufp = cmdline; *bufp; ) {
		/* Skip leading whitespace */
		while ( isspace(*bufp) ) {
			++bufp;
		}
		/* Skip over argument */
		if ( *bufp == '"' ) {
			++bufp;
			if ( *bufp ) {
				if ( argv ) {
					argv[argc] = bufp;
				}
				++argc;
			}
			/* Skip over word */
			while ( *bufp && (*bufp != '"') ) {
				++bufp;
			}
		} else {
			if ( *bufp ) {
				if ( argv ) {
					argv[argc] = bufp;
				}
				++argc;
			}
			/* Skip over word */
			while ( *bufp && ! isspace(*bufp) ) {
				++bufp;
			}
		}
		if ( *bufp ) {
			if ( argv ) {
				*bufp = '\0';
			}
			++bufp;
		}
	}
	if ( argv ) {
		argv[argc] = NULL;
	}
	return(argc);
}

/* This is where execution begins [console apps] */
int console_main(int argc, char *argv[])
{
    int iRet = 0;
    
	/* Run the application main() code */
	iRet = mg_main(argc, argv);

	return(0);
}

int WINAPI WinMain(
    HINSTANCE hInstance,      /* handle to current instance */
    HINSTANCE hPrevInstance,  /* handle to previous instance */
    LPSTR lpCmdLine,          /* command line */
    int nCmdShow              /* show state */
)
{
	char **argv;
	int argc;

  	argc = ParseCommandLine__(lpCmdLine, NULL);
	argv = (char **)alloca((argc+1)*(sizeof *argv));

	/* Run the main program (after a little SDL initialization) */
	return(console_main(argc, argv));
}
/*
 * ----------------------------------------------------------------------------
 * global functions
 * ----------------------------------------------------------------------------
 */

void app_set_win32_instance(HINSTANCE instance)
{
	mghInstance = instance;
}

HINSTANCE app_get_win32_instance()
{
    return mghInstance;
}
