
#include <stdio.h>

#include "SDL_thread.h"

#include "mg.h"
#include "mg_misc.h"
#include "mg_gal.h"
#include "mg_cliprect.h"
#include "mg_internals.h"
#include "mg_event.h"
#include "mg_cursor.h"
#include "mg_ctrlclass.h"
#include "mg_accelkey.h"
#include "mg_timer.h"

/************************** System Initialization ****************************/
static BOOL InitSystemThreads(void)
{
    return TRUE;
}

int  mg_InitSubSystem()
{
    // Init miscelleous
    if (!InitMisc ()) {
        fprintf (stderr, "DESKTOP: Initialization of misc things failure!\n");
        return FALSE;
    }

    switch (InitGAL ()) {
    case ERR_CONFIG_FILE:
        fprintf (stderr, 
            "GDI: Reading configuration failure!\n");
        return -1;

    case ERR_NO_ENGINE:
        fprintf (stderr, 
            "GDI: No graphics engine defined!\n");
        return -1;

    case ERR_NO_MATCH:
        fprintf (stderr, 
            "GDI: Can not get graphics engine information!\n");
        return -1;

    case ERR_GFX_ENGINE:
        fprintf (stderr, 
            "GDI: Can not initialize graphics engine!\n");
        return -1;
    }

    // same : sdl_fatal.c , Parachute using signal
#if 0 // windows
    InstallSEGVHandler ();
#endif

    /* Init GDI. */
    if(!InitGDI()) {
        fprintf (stderr, "DESKTOP: Initialization of GDI failure!\n");
        goto failure1;
    }

#ifdef _USE_NEWGAL
    /* Init Screen DC here */
    if (!InitScreenDC ()) {
        fprintf (stderr, "Can not initialize screen DC!\n");
        goto failure1;
    }
#endif

    if (!InitWindowElementColors ()) {
        fprintf (stderr, "DESKTOP: Can not initialize colors of window element!\n");
        goto failure1;
    }

    // Init low level event
    if(!InitLWEvent()) {
        fprintf(stderr, "DESKTOP: Low level event initialization failure!\n");
        goto failure1;
    }

    if (!InitFixStr ()) {
        fprintf (stderr, "DESKTOP: Init Fixed String module failure!\n");
        goto failure2;
    }
    
#ifdef _CURSOR_SUPPORT
    // Init mouse cursor.
    if( !InitCursor() ) {
        fprintf (stderr, "DESKTOP: Count not initialize mouse cursor support!\n");
        goto failure2;
    }
#endif

    // Init menu
    if (!InitMenu ()) {
        fprintf (stderr, "DESKTOP: Init Menu module failure!\n");
        goto failure2;
    }

    // Init control class
    if(!InitControlClass()) {
        fprintf(stderr, "DESKTOP: Init Control Class failure!\n");
        goto failure2;
    }

    // Init accelerator
    if(!InitAccel()) {
        fprintf(stderr, "DESKTOP: Init Accelerator failure!\n");
        goto failure2;
    }

    if (!InitSystemThreads())
        goto failure2;

    return 0;
	
failure2:
	    TerminateLWEvent ();
failure1:
	    TerminateGAL ();
	return -1;
}

void mg_TerminateSubSystem()
{
    TerminateTimer ();
    TerminateDesktop ();
    TerminateAccel ();
    TerminateControlClass ();
    TerminateMenu ();
#ifdef _CURSOR_SUPPORT
    TerminateCursor ();
#endif
    TerminateFixStr ();
    TerminateLWEvent ();

#ifdef _USE_NEWGAL
    TerminateScreenDC ();
#endif

    TerminateGDI ();
    TerminateGAL ();
    TerminateMisc ();
}

// move from init.c
BOOL InitGUI (void)
{
	/* Clear the error message */
	//mg_ClearError();

#if !defined(_WINDOWS)
#if defined(__NOLINUX__) 
    // Save original termio
    tcgetattr (0, &savedtermio);
#endif /* !defined(_WINDOWS) */
#endif /* !defined(_WINDOWS) */
    
	/* Initialize the desired subsystems */
	if ( mg_InitSubSystem() < 0 ) {
		return FALSE;
	}

    SetKeyboardLayout ("default");

    SetCursor (GetSystemCursor (IDC_ARROW));
    SetCursorPos (g_rcScr.right >> 1, g_rcScr.bottom >> 1);

    TerminateMgEtc ();
	

	return TRUE;
}

// move from init.c
void TerminateGUI(int rcByGUI)
{
#if defined(__WINDOWS__)
    if (rcByGUI >= 0) {
        SDL_WaitThread(__mg_timer, NULL);
    }
#elif defined(__ECOS)
    if (rcByGUI >= 0) {
        pthread_join (__mg_timer, NULL);
    }
#else
    if (rcByGUI >= 0) {
        pthread_join (__mg_timer, NULL);
    }
    else {
        pthread_kill_other_threads_np ();
    }
#endif /* defined(__WINDOWS__) */

    DestroyMsgQueue (&__mg_dsk_msgs);

    DestroyFreeQMSGList ();

    mg_TerminateSubSystem();
}

#if defined(_WIN32_WCE) || (defined(__WATCOMC__) && defined(BUILD_DLL)) 
/* Need to include DllMain() on Windows CE and Watcom C for some reason.. */
#include <windows.h>

BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
    switch (ul_reason_for_call)
	{
		case DLL_PROCESS_ATTACH:
		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
    }
    return TRUE;
}
#endif 
