/*
** $Id: ial_native.c,v 1.1 2004/12/17 04:40:42 heavenzen Exp $
**
** native.c: Native IAL engine
**
** Copyright (C) 2003 Feynman Software.
** Copyright (C) 2000 ~ 2002 Song Lixin and Wei Yongming.
**
** Created by Song Lixin at 2000/10/17
** Clean code for MiniGUI 1.1.x by Wei Yongming, 20001/09/18
*/

/*
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#define strncasecmp strncmp
#ifndef _WINDOWS
#include <sys/ioctl.h>
#endif
#include <signal.h>
#ifndef _WINDOWS
#include <linux/kd.h>
#include <linux/keyboard.h>
#endif

#include "common.h"
#include "minigui.h"
#include "misc.h"
#include "ial.h"
#include "gal.h"
#include "ial_native.h"

#define FD_ZERO(a)				1
#define FD_SET(a,b)				1
#define select(a,b,c,d,e)		1


static int mouse_fd;
static int kbd_fd;

static int    xpos;        /* current x position of mouse */
static int    ypos;        /* current y position of mouse */
static int    minx;        /* minimum allowed x position */
static int    maxx;        /* maximum allowed x position */
static int    miny;        /* minimum allowed y position */
static int    maxy;        /* maximum allowed y position */
static int    buttons;     /* current state of buttons */
static int    scale;       /* acceleration scale factor */
static int    thresh;      /* acceleration threshhold */

static MOUSEDEVICE * mousedev;

static unsigned char state [NR_KEYS];
static KBDDEVICE * kbddev = &kbddev_tty;

/************************  Low Level Input Operations **********************/

/*
 * Mouse operations -- Event
 */
static int mouse_update(void)
{
    int dx,dy,dz;
    int r;
    int sign;
    
    if (!mousedev)
        return 0;

    dx = dy = 0;
    r = mousedev->Read(&dx, &dy, &dz, &buttons);
    if (r <= 0)
        return 0;

    sign = 1;
    if (dx < 0) {
        sign = -1;
        dx = -dx;
    }
    if (dx > thresh)
        dx = thresh + (dx - thresh) * scale;
    dx *= sign;
#ifdef _WINDOWS
	xpos = dx;
#else
    xpos += dx;
#endif	//_WINDOWS
    if( xpos < minx )
        xpos = minx;
    if( xpos > maxx )
        xpos = maxx;

    sign = 1;
    if (dy < 0) {
        sign = -1;
        dy = -dy;
    }
    if (dy > thresh)
        dy = thresh + (dy - thresh) * scale;
    dy *= sign;
#ifdef _WINDOWS
	ypos = dy;
#else
    ypos += dy;
#endif	//_WINDOWS
    if ( ypos < miny )
        ypos = miny;
    if ( ypos > maxy )
        ypos = maxy;

    return r;
}

static void mouse_getxy (int* x, int* y)
{
#ifdef _COOR_TRANS
#if _ROT_DIR_CCW
    *x = ypos;
    *y = maxx - xpos;
#else
    *x = maxy - ypos;
    *y = xpos;
#endif
#else
    *x = xpos;
    *y = ypos;
#endif
}

static void mouse_setposition (int newx, int newy)
{
#ifdef _COOR_TRANS
#if _ROT_DIR_CCW
    int tmp = newx;
    newx = maxx - newy;
    newy = tmp;
#else
    int tmp = newy;
    newy = maxy - newx;
    newx = tmp;
#endif
#endif
    if (newx < minx)
        newx = minx;
    if (newx > maxx)
        newx = maxx;
    if (newy < miny)
        newy = miny;
    if (newy > maxy)
        newy = maxy;
    if (newx == xpos && newy == ypos)
        return;
    xpos = newx;
    ypos = newy;
}

static int mouse_getbutton (void)
{
    return buttons;
}

static void mouse_setrange (int newminx, int newminy, int newmaxx, int newmaxy)
{
#ifdef _COOR_TRANS
    return;
#else
    minx = newminx;
    miny = newminy;
    maxx = newmaxx;
    maxy = newmaxy;
    mouse_setposition ((newminx + newmaxx) / 2, (newminy + newmaxy) / 2);
#endif
}

static void mouse_suspend(void)
{
    if (mousedev) {
        mousedev->Suspend();
        mouse_fd = -1;
    }
}

static int mouse_resume(void)
{
    if (mousedev) {
        mouse_fd = mousedev->Resume();
        if (mouse_fd < 0) {
            fprintf (stderr, "IAL Native Engine: Can not open mouse!\n");
        }
    }

    return mouse_fd;
}
/*
 * 0 : no data
 * > 0 : the possible maximal key scancode
 */
static int keyboard_update(void)
{
    unsigned char buf;
    int modifier;
    int ch;
    int is_pressed;
    int retvalue;

    retvalue = kbddev->Read (&buf, &modifier); 

    if ((retvalue == -1) || (retvalue == 0))
        return 0;

    else { /* retvalue > 0 */
        is_pressed = !(buf & 0x80);
        ch         = buf & 0x7f;
#ifdef KEYBOARD_DEBUG
{
		FILE *fp;
		fp = fopen("presskey.dat","a+");
		fprintf(fp, "int keyboard_update():is_pressed=%d, ch=%d, \n", is_pressed, ch);
		fclose(fp);
}
#endif

        if (is_pressed) {
#ifdef _LITE_VERSION
            if ((ch >= SCANCODE_F1) && (ch <= SCANCODE_F6) && 
                            (state[SCANCODE_RIGHTCONTROL]==1)) 
            {
                vtswitch_try (ch - SCANCODE_F1 + 1);
                state [SCANCODE_RIGHTCONTROL] = 0;
                return 0;
            }
#endif
            state[ch] = 1;
        }
        else 
            state[ch] = 0;
    }

    return NR_KEYS;
}

static const char* keyboard_getstate(void)
{
    return state;
}

static void keyboard_suspend(void)
{
    kbddev->Suspend();    
    kbd_fd = -1;
}

static int keyboard_resume(void)
{
    memset(state, 0, NR_KEYS);
    kbd_fd = kbddev->Resume ();
    if (kbd_fd < 0) {
        fprintf (stderr, "IAL Native Engine: Can not open keyboard!\n");
    }

    return kbd_fd;
}

/* NOTE by weiym: Do not ignore the fd_set in, out, and except */
#ifdef _LITE_VERSION
static int wait_event (int which, int maxfd, fd_set *in, fd_set *out, fd_set *except,
                struct timeval *timeout)
#else
static int wait_event (int which, fd_set *in, fd_set *out, fd_set *except,
                struct timeval *timeout)
#endif
{
    fd_set    rfds;
    int    retvalue = 0;
    int    fd, e;

    if (!in) {
        in = &rfds;
        FD_ZERO (in);
    }

    if (which & IAL_MOUSEEVENT && mouse_fd >= 0) {
        fd = mouse_fd;          /* FIXME: mouse fd may be changed in vt switch! */
        FD_SET (fd, in);
#ifdef _LITE_VERSION
        if (fd > maxfd) maxfd = fd;
#endif
    }

    if (which & IAL_KEYEVENT && kbd_fd >= 0){
        fd = kbd_fd;          /* FIXME: keyboard fd may be changed in vt switch! */
        FD_SET (kbd_fd, in);
#ifdef _LITE_VERSION
        if (fd > maxfd) maxfd = fd;
#endif
    }

    /* FIXME: pass the real set size */
#ifdef _LITE_VERSION
    e = select (maxfd + 1, in, out, except, timeout) ;
#else
    e = select (FD_SETSIZE, in, out, except, timeout) ;
#endif
#ifdef _WINDOWS
	while(1)
	{
		// mousekey_sem_wait(); external fuction
		if (mousedev_PS2.Resume())
			retvalue |= IAL_MOUSEEVENT;
		if (kbddev_tty.Resume())
			retvalue |= IAL_KEYEVENT;
		if (retvalue != 0)
			return retvalue;
	}
#else
    if (e > 0) 
{ 
        fd = mouse_fd;
        /* If data is present on the mouse fd, service it: */
        if (fd >= 0 && FD_ISSET (fd, in)) {
            FD_CLR (fd, in);
            retvalue |= IAL_MOUSEEVENT;
        }

        fd = kbd_fd;
        /* If data is present on the keyboard fd, service it: */
        if (fd >= 0 && FD_ISSET (fd, in)) {
            FD_CLR (fd, in);
            retvalue |= IAL_KEYEVENT;
        }
    } 
else if (e < 0) 
{
        return -1;
    }
#endif // _WINDOWS
    return retvalue;
}

static void set_leds (unsigned int leds)
{
#ifndef _WINDOWS
    ioctl (0, KDSETLED, leds);
#else 
/* Code Added by jx, bit0:scrolllock,bit1:numlock,bit2:capslock*/
	extern void win_kbd_set_leds(unsigned int leds);
	win_kbd_set_leds(leds);
#endif	//_WINDOWS
}

BOOL InitNativeInput (INPUT* input, const char* mdev, const char* mtype)
{
    kbd_fd = kbddev->Open();    
    if (kbd_fd < 0) {
        perror ("IAL Native Engine");
        return FALSE;
    }
    
    if (strncasecmp (mtype, "none", 4) == 0) {
        mousedev = NULL;
        goto found;
    }
#ifdef _IMPS2_SUPPORT
    if (strncasecmp (mtype, "imps2", 5) == 0 ) {
        mousedev = &mousedev_IMPS2;
        goto found;
    }
#endif
#ifdef _PS2_SUPPORT
    if (strncasecmp (mtype, "ps2", 3) == 0 ) {
        mousedev = &mousedev_PS2;
        goto found;
    }
#endif
#ifdef _GPM_SUPPORT
    if (strncasecmp (mtype, "gpm", 3) == 0) {
        mousedev = &mousedev_GPM;
        goto found;
    }
#endif
#ifdef _MS_SUPPORT
    if (strncasecmp (mtype, "ms", 3) == 0) {
        mousedev = &mousedev_MS;
        goto found;
    }
#endif
#ifdef _MS3_SUPPORT
    if (strncasecmp (mtype, "ms3", 3) == 0) {
        mousedev = &mousedev_MS3;
        goto found;
    }
#endif

    fprintf (stderr, "IAL Native Engine: Can not init pointing device!\n");
    return FALSE;

found:
    if (mousedev) {
        mouse_fd = mousedev->Open(mdev);
        if (mouse_fd <0) {
            fprintf (stderr, "IAL Native Engine: Can not init mouse!\n");
            kbddev->Close();
            return FALSE;
        }
    }
    else
        mouse_fd = -1;

    xpos = 0;
    ypos = 0;
    buttons = 0;
    minx = 0;
    miny = 0;
#ifdef _COOR_TRANS
    maxx = HEIGHTOFPHYGC - 1;
    maxy = WIDTHOFPHYGC - 1;
#else
    maxx = WIDTHOFPHYGC - 1;
    maxy = HEIGHTOFPHYGC - 1;
#endif

    if (mousedev)
        mousedev->GetDefaultAccel (&scale, &thresh);
    input->update_mouse = mouse_update;
    input->get_mouse_xy = mouse_getxy;
    input->set_mouse_xy = mouse_setposition;
    input->get_mouse_button = mouse_getbutton;
    input->set_mouse_range = mouse_setrange;
    input->suspend_mouse= mouse_suspend;
    input->resume_mouse = mouse_resume;

    input->update_keyboard = keyboard_update;
    input->get_keyboard_state = keyboard_getstate;
    input->suspend_keyboard = keyboard_suspend;
    input->resume_keyboard = keyboard_resume;
    input->set_leds = set_leds;

    input->wait_event = wait_event;
    
#ifdef _LITE_VERSION
#ifndef _STAND_ALONE
    if (mgIsServer)
#endif
        init_vtswitch (kbd_fd);
#endif

    return TRUE;
}

void TermNativeInput (void)
{
#ifdef _LITE_VERSION
#ifndef _STAND_ALONE
    if (mgIsServer)
#endif
        done_vtswitch (kbd_fd);
#endif

    if (mousedev)
        mousedev->Close();
    kbddev->Close();
}

